#include <xc.h>

// コンフィギュレーションビット設定
#pragma config FOSC = INTOSC    // 内部OSCを使用
#pragma config WDTE = OFF       // ウォッチドッグタイマ無効
#pragma config LVP = ON         // 低電圧プログラミング有効
#pragma config MCLRE = OFF      // MCLRピン機能無効

#define _XTAL_FREQ 32000000     // 内蔵OSCの周波数設定（32MHz）

int Kp = 2;                     // 比例ゲイン
int Ki = 1;                     // 積分ゲイン
int integral = 0;               // 積分値
int currentSpeed = 0;           // 現在の速度（フィードバック）
int targetSpeed = 500;          // 目標速度（例）

// 上下限値の設定
const int integral_max = 255;
const int integral_min = -255;
const int controlSignal_max = 249; // PR2と同じ値に設定
const int controlSignal_min = 0;

int controlSignal; // 制御信号

void setupADC(void) {
    // ADCの設定
    ADCON0 = 0; // ADCをオフにして初期化
    ADCON1 = 0b11100000; // 右詰め結果、Fosc/32
    ADCON0bits.ADON = 1; // ADCを有効化
}

void selectChannel(unsigned char channel) {
    // ADCチャンネルを選択
    ADCON0 &= 0b11000011; // チャンネル選択をリセット
    ADCON0 |= (channel << 2); // 新しいチャンネルを設定
}

unsigned int readADC(void) {
    // ADC変換の開始
    ADCON0bits.GO = 1;
    // 変換の完了を待つ
    while(ADCON0bits.GO_nDONE);
    // 結果を返す (右詰めのため、ADRESの値をそのまま返す)
    return ADRES;
}

void setup(void) {
    // クロックの初期設定
    OSCCON = 0b11110000; // IRCF = 1111 (32MHz), SCS = 00 (クロック選択はコンフィグワードに依存)
    
    // タイマ1設定
    T1CON = 0b00110001; // 1:1プリスケーラ, FOSC/4, タイマ1 ON
    TMR1H = (65536 - 1000) >> 8; 	// PWM周期を設定（1ms）
    TMR1L = (65536 - 1000) & 0xFF; 	// PWM周期を設定（1ms）

    // PWM設定（CCP2）
    CCP2CON = 0b10111100; // ハーフブリッジモード設定
    APFCON1bits.CCP2SEL = 1; // CCP2(P2A)をRA4に割り当て
    APFCON1bits.P2BSEL = 1; // P2B(PWM出力ピン)をRA5に割り当て
    TRISA4 = 0; // RA4(P2B)ピンを出力に設定
    TRISA5 = 0; // RA5(P2A)ピンを出力に設定
    PR2 = 249; // PWM周期（32kHzにするための設定）
    T2CON = 0b00000100; // タイマ2設定（プリスケーラ1:4、タイマ2 ON）
    CCPR2L = 0; // PWMデューティ比初期値（0%）
    
    // RC0ピンをデジタル出力に設定
    TRISC0 = 0;
    
    // ADCの初期設定を追加
    setupADC();
    
    // 割り込み設定
    PIE1bits.TMR1IE = 1; // タイマ1割り込み有効
    INTCONbits.PEIE = 1; // ペリフェラル割り込み有効
    INTCONbits.GIE = 1;  // グローバル割り込み有効
}

void updateMotorSpeed(void) {
    // AN2から速度フィードバック値を読み取り
    selectChannel(2); // AN2を選択
    currentSpeed = readADC(); // 現在の速度をADC値として読み取る
    
    // AN6からモータ速度設定値を読み取り
    selectChannel(6); // AN6を選択
    targetSpeed = readADC(); // 目標速度をADC値として読み取る
    
    // PI制御の計算
    int error = targetSpeed - currentSpeed; // 誤差
    int Pout = Kp * error; // 比例項
    integral += error; // 積分項
    // 上下限値のチェック
    if (integral > integral_max) integral = integral_max;
    if (integral < integral_min) integral = integral_min;
    
    int Iout = Ki * integral;
    controlSignal = Pout + Iout; // PI制御出力
    // 上下限値のチェック
    if (controlSignal > controlSignal_max) controlSignal = controlSignal_max;
    if (controlSignal < controlSignal_min) controlSignal = controlSignal_min;
    
    // PWMデューティ比を設定
    CCPR2L = (unsigned char)(controlSignal); // デューティ比を設定
}

void __interrupt() isr(void) {
    static int toggle = 0; // RC0のトグル用変数

    RC0 = 1;
    
    if (PIR1bits.TMR1IF) { // タイマ1割り込みフラグチェック
        PIR1bits.TMR1IF = 0; // タイマ1割り込みフラグをクリア

        TMR1H = (65536 - 1000) >> 8; 	// PWM周期を設定（1ms）
        TMR1L = (65536 - 1000) & 0xFF; 	// PWM周期を設定（1ms）


        updateMotorSpeed(); // モータ速度の更新
    }

    RC0 = 0;
}

void main(void) {
    setup();
    
    while(1) {
        // メインループで行う処理
        // この例では、すべての処理は割り込みによって行われるため、ここは空です。
    }
}
